/******************************************************************************
      Image Acquisition & Display Program - Primitive Control Functions
                       Runs on any VGA or MCGA display.
 Displays the image with the serial register on the right edge of the screen.
 Operates two cameras simultaneously and displays two picture upon keyboard
 entry.
NOTE - Does not use a trigger exposes on camera-op execute. Takes only one
       picture.
                                         Kev Forrester Jan 14 92
******************************************************************************/
#include <bios.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <at200.h>
#include <at200oem.h>
#include <camera.h>

#define DEFAULT_CCD      4            /* Thomson 512 x 512 CCD */
#define DEFAULT_ANALOG   A200KHZ      /* 200KHz 12 bit analog electronics */
#define DEFAULT_EXPOSURE 2            /* Exposure time in milliseconds */
#define DEFAULT_SHUTTER  40            /* Shutter open/close time in milliseconds */

int ccd_type = DEFAULT_CCD,
    ser_byte = DEFAULT_ANALOG,
    exposure = DEFAULT_EXPOSURE,
    shutter  = DEFAULT_SHUTTER,
    bin_ccd  = 0;                     /* 0=Bin 1x1, Non-Zero=Bin MAXxMAX */

char far *screen_row[200] = { 0, };   /* Pointers to graphics display rows */
/* screen_row[] is initialized in main() to point to the start of each row on
   the VGA display.  It is then used by the show_image() function to update
   the VGA memory. */

unsigned huge buffer[65536] = { 0, }, /* Camera Image Buffer */
         far *ptr[320] = { 0, },      /* Image Display Pointers */
         far *ptable[32] = { 0,};     /* Image Collection Pointers */
/* buffer[] is an array of 64 K pixels (128 K bytes) that is used to collect
   data from the camera.  The ptable[] array is used by the AT200 control
   library functions to collect the pixels.  It is initialized in main() to
   point to each 2 K pixel (4 K byte) block in the buffer[] array.  The ptr[]
   array is used by show_image() to display 320 pixel rows on the VGA.  It is
   also initialized in main(). */


void mcga_display_mode(int mode)
{ /* Select text mode or 320 x 200 - 256 color graphics mode */
union REGS regs;

  regs.x.ax = (mode) ? 19 : 3;  /* AH=0, AL=Mode */
  int86(16, &regs, &regs);      /* Call video bios */
}

void mcga_color_set(int color,int red,int green,int blue)
{ /* Set the RGB values for a color in 256 color graphics modes */
  outp(0x3C8,color);  /* DAC color control register selects the color */
  outp(0x3C9,red);    /* Set Red DAC (0..63) */
  outp(0x3C9,green);  /* Set Green DAC (0..63) */
  outp(0x3C9,blue);   /* Set Blue DAC (0..63) */
}

void far *normalize_pointer(void far *data)
/* Make pointers have the lowest offset possible (and the largest segment).
   This function is needed to insure that each pointer in the ptable[] array
   can be used for 2 K pixels (4 K bytes).  Taking the address of an element
   of a huge array only guarantees that the segment:offset pair will be valid
   for that element.  The calculation performed here guarantees that the
   segment:offset pair will be valid for at least the 4 K addresses we need.
   See the Microsoft Macro Assembler 5.1 manual, section 13.2 for more
   information on REAL MODE ADDRESSING. */
{
unsigned seg,offs;

  seg  = (unsigned) ((long) data >> 16L);
  offs = (unsigned) ((long) data & 0xFFFF);
  seg += offs >> 4;
  offs &= 15;
  return( (void far *) (((long) seg << 16L) | (long) offs) );
}

void load_ops(void)
{
unsigned x,p_bin,s_bin,rows,cols;

  if( bin_ccd ) {
    p_bin = ccd[ccd_type].rows / 320;
    if( p_bin < 1 )
      p_bin = 1;
    s_bin = ccd[ccd_type].cols / 200;
    if( s_bin < 1 )
      s_bin = 1;
  }
  else
    p_bin = s_bin = 1;
  rows = ccd[ccd_type].rows / p_bin;
  cols = ccd[ccd_type].cols / s_bin;
  disable_trigger();
  outp(at200_base + SER_CONTROL,ser_byte);
  x = at200_base + CONTROL;
  outp(x,inp(x) & ~FIFO_INPUT_SELECT);
  dsp_command( SET_PARALLEL_DELAY );
  dsp_arg( (long) ccd[ccd_type].parallel_delay );
  dsp_command( SET_PARALLEL_STATES );
  dsp_arg( (long) ccd[ccd_type].parallel_states );
  dsp_command( SET_SHUTTER_CLOSE );
  dsp_arg( (long) shutter );
  dsp_command( SET_SHUTTER_OPEN );
  dsp_arg( (long) shutter );
  dsp_command( LOAD_OPERATIONS );
  dsp_op( OP_SET_GAIN,1 );
  dsp_op( OP_SET_SER_BIN,s_bin );
  dsp_op( OP_EXPOSE_SCALE,exposure ); /* Exposure Timer Scale in msec */
  dsp_op( OP_PARALLEL_DIS,ccd[ccd_type].rows );

    dsp_op( OP_OPEN_SHUTTER,0 );
    dsp_op( OP_EXPOSE,1 );            /* Expose 1 * Timer Scale */
    dsp_op( OP_CLOSE_SHUTTER,0 );
  dsp_op( OP_TERMINATE,0 );
  for( x=1 ; x &&
     (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) ; x++ )
    ; /* Wait for DSP to complete initialization */
  dsp_command( EXECUTE );
}

void load_ops2(void)
{
unsigned x,p_bin,s_bin,rows,cols;

  if( bin_ccd ) {
    p_bin = ccd[ccd_type].rows / 320;
    if( p_bin < 1 )
      p_bin = 1;
    s_bin = ccd[ccd_type].cols / 200;
    if( s_bin < 1 )
      s_bin = 1;
  }
  else
    p_bin = s_bin = 1;
  rows = ccd[ccd_type].rows / p_bin;
  cols = ccd[ccd_type].cols / s_bin;
  dsp_command( LOAD_OPERATIONS );
    x = (ccd[ccd_type].rows - 320 * p_bin) / 2;
    if( x )
      dsp_op( OP_PARALLEL_DIS,x);
    x = ccd[ccd_type].cols + ccd[ccd_type].prescan + ccd[ccd_type].postscan;
    dsp_op( OP_SERIAL_DIS,x * 5);     /* Clear Serial Register */
    dsp_op( OP_DO_LOOP,320 );
      dsp_op( OP_PARALLEL_BIN,p_bin );
      x = (ccd[ccd_type].cols - 200 * s_bin) / 2 + ccd[ccd_type].prescan;
      if( x )
        dsp_op( OP_SERIAL_DIS,x);
      dsp_op( OP_SERIAL_CONV,200 );
      x += 200 * s_bin;
      if( ccd[ccd_type].postscan > ccd[ccd_type].prescan )
        x -= ccd[ccd_type].postscan - ccd[ccd_type].prescan;
      dsp_op( OP_SERIAL_DIS,ccd[ccd_type].cols + ccd[ccd_type].prescan - x);
    dsp_op( OP_END_DO,0 );
    x = (ccd[ccd_type].rows - 320 * p_bin) / 2;
    if( x )
      dsp_op( OP_PARALLEL_DIS,x);
  dsp_op( OP_TERMINATE,0 );
  for( x=1 ; x &&
     (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) ; x++ )
    ; /* Wait for DSP to complete initialization */
  dsp_command( EXECUTE );
}

void show_image(void)
{
unsigned char buff[320];
unsigned max,min;
register int x,y;

  reset_fifo();
  load_ops2();
  while( !(inp(at200_base + STATUS) & FIFO_EMPTY) )  /* Wait for 1st Pixel */
  ;
  collect_image(ptable);    /* Use the fast routine for up to 64K Pixels */
  if( !(inp(at200_base + STATUS) & FIFO_OVERFLOW) ) /* Check for overflow */
    return;
  for( min=65535,max=y=0 ; y<200 ; y++ )
    for( x=319 ; x>=0 ; x-- ) {
      if( max < ptr[x][y] )
        max = ptr[x][y];
      if( min > ptr[x][y] )
        min = ptr[x][y];
    }
  max = (unsigned) (((long) max - (long) min + 62L) / 63L);
  if( !max )
    max = 1;    /* Don't divide by zero */
  for( y=0 ; y<200 ; y++ ) {
    for( x=319 ; x>=0 ; x-- )
      buff[319-x] = (unsigned char) ((ptr[x][y] - min) / max);
    memcpy(screen_row[y],buff,320);
  }
}


void main()
{
register unsigned x;
int data_bits = 14;
      ser_byte = 32;

/* Load setup program into camera 2 */
        at200_base = 0x230;
        x = load_dsp_program(dsp_cam200k,size_of_dsp_cam200k);

  printf("KHz Analog Card.\n");
  if( !x ) {
    printf("Error loading program.\n");
    exit( 1 );
  }
  for( x=1 ; x ; x++ )
    if( !(inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) )
      break;  /* Wait for DSP to complete initialization */
  if( !x ) {  /* Timed out! */
    printf("The AT200 is unresponsive or missing.  Unable to proceed.\n");
    exit( 2 );
  }
  for( x=0 ; x<32 ; x++ )     /* Load image acquisition array */
    ptable[x] = normalize_pointer(&buffer[x * 2048]);
  for( x=0 ; x<320 ; x++ )    /* Load image display array */
    ptr[x] = normalize_pointer(&buffer[x * 200]);
  for( x=0 ; x<200 ; x++ )    /* Load graphics display array */
    screen_row[x] = &((char far *)0xA0000000)[x * 320];

/* Load operations into camera 2 */
  load_ops();

/* Load setup program into camera 1 */

        at200_base = 0x220;
        x = load_dsp_program(dsp_cam200k,size_of_dsp_cam200k);

  printf("KHz Analog Card.\n");
  if( !x ) {
    printf("Error loading program.\n");
    exit( 1 );
  }
  for( x=1 ; x ; x++ )
    if( !(inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) )
      break;  /* Wait for DSP to complete initialization */
  if( !x ) {  /* Timed out! */
    printf("The AT200 is unresponsive or missing.  Unable to proceed.\n");
    exit( 2 );
  }

/* Load operations into camera 1 */
  load_ops();
  x=getchar();  /* Need this or first image gets blurred. Don't ask me why. */

/* Set up VGA */
  mcga_display_mode( 1 );
  for( x=0 ; x<64 ; x++ )
    mcga_color_set(x,x,x,x);

/* Display image of camera 1 */
    show_image();
    x=getchar();


/* Display image of camera 2 */
    at200_base = 0x230;

    show_image();
    x=getchar();

  mcga_display_mode( 0 );
  exit( 0 );
}
